# FROM HF TRL

# Copyright 2024 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing as tp

from easydel.infra.utils import ProcessingClassType

DatasetType = tp.TypeVar("DatasetType")


def is_conversational(example: tp.Dict[str, tp.Any]) -> bool:
	r"""
	Check if the example is in a conversational format.

	Args:
	    example (`tp.Dict[str, tp.Any]`):
	        A single data entry of a dataset. The example can have different keys depending on the
	        dataset type.

	Returns:
	    `bool`: `True` if the data is in a conversational format, `False` otherwise.

	Examples:

	>>> example = {"prompt": [{"role": "user", "content": "What color is the sky?"}]}
	>>> is_conversational(example)
	... True
	>>> example = {"prompt": "The sky is"})
	>>> is_conversational(example)
	... False
	```
	"""
	supported_keys = ["prompt", "chosen", "rejected", "completion", "messages"]
	example_keys = {key for key in example.keys() if key in supported_keys}

	if example_keys:
		key = example_keys.pop()
		maybe_messages = example[key]
		if isinstance(maybe_messages, list):
			maybe_message = maybe_messages[0]

			if (
				isinstance(maybe_message, dict)
				and "role" in maybe_message
				and "content" in maybe_message
			):
				return True

	return False


def apply_chat_template(
	example: tp.Dict[str, tp.List[tp.Dict[str, str]]],
	processing_class: ProcessingClassType,
) -> tp.Dict[str, str]:
	r"""
	Apply a chat template to a conversational example.

	For more details, see [`maybe_apply_chat_template`].
	"""
	# Check that the example has the correct keys
	supported_keys = ["prompt", "chosen", "rejected", "completion", "messages", "label"]
	example_keys = {key for key in example.keys() if key in supported_keys}
	if example_keys not in [
		{"messages"},  # language modeling
		{"prompt"},  # prompt-only
		{"prompt", "completion"},  # prompt-completion
		{"prompt", "chosen", "rejected"},  # preference
		{"chosen", "rejected"},  # preference with implicit prompt
		{"prompt", "completion", "label"},  # unpaired preference
	]:
		raise KeyError(f"Invalid keys in the example: {example_keys}")

	# Apply the chat template to the whole conversation
	if "messages" in example:
		messages = processing_class.apply_chat_template(example["messages"], tokenize=False)

	# Apply the chat template to the prompt, adding the generation prompt
	if "prompt" in example:
		prompt = processing_class.apply_chat_template(
			example["prompt"], tokenize=False, add_generation_prompt=True
		)

	# Apply the chat template to the entire prompt + completion
	if "prompt" in example:  # explicit prompt and prompt-completion case
		if "chosen" in example:
			prompt_chosen = processing_class.apply_chat_template(
				example["prompt"] + example["chosen"], tokenize=False
			)
			chosen = prompt_chosen[len(prompt) :]
		if "rejected" in example and "prompt" in example:  # explicit prompt
			prompt_rejected = processing_class.apply_chat_template(
				example["prompt"] + example["rejected"], tokenize=False
			)
			rejected = prompt_rejected[len(prompt) :]
		if "completion" in example:
			prompt_completion = processing_class.apply_chat_template(
				example["prompt"] + example["completion"], tokenize=False
			)
			completion = prompt_completion[len(prompt) :]
	else:  # implicit prompt case
		if "chosen" in example:
			chosen = processing_class.apply_chat_template(example["chosen"], tokenize=False)
		if "rejected" in example:
			rejected = processing_class.apply_chat_template(
				example["rejected"], tokenize=False
			)

	# Ensure that the prompt is the initial part of the prompt-completion string
	if "prompt" in example:
		error_message = (
			"The chat template applied to the prompt + completion does not start with the chat template applied to "
			"the prompt alone. This can indicate that the chat template is not supported by TRL."
			"\n**Prompt**:\n{}\n\n**Prompt + Completion**:\n{}"
		)
		if "chosen" in example and not prompt_chosen.startswith(prompt):
			raise ValueError(error_message.format(prompt, prompt_chosen))
		if "rejected" in example and not prompt_rejected.startswith(prompt):
			raise ValueError(error_message.format(prompt, prompt_rejected))
		if "completion" in example and not prompt_completion.startswith(prompt):
			raise ValueError(error_message.format(prompt, prompt_completion))

	# Extract the completion by removing the prompt part from the prompt-completion string
	output = {}
	if "messages" in example:
		output["text"] = messages
	if "prompt" in example:
		output["prompt"] = prompt
	if "chosen" in example:
		output["chosen"] = chosen
	if "rejected" in example:
		output["rejected"] = rejected
	if "completion" in example:
		output["completion"] = completion
	if "label" in example:
		output["label"] = example["label"]

	return output


def maybe_apply_chat_template(
	example: tp.Dict[str, tp.List[tp.Dict[str, str]]],
	processing_class: ProcessingClassType,
) -> tp.Dict[str, str]:
	r"""
	If the example is in a conversational format, apply a chat template to it.

	Args:
	    example (`tp.Dict[str, tp.List[tp.Dict[str, str]]`):
	        Dictionary representing a single data entry of a conversational dataset. Each data entry can have different
	        keys depending on the dataset type. The supported dataset types are:

	            - Language modeling dataset: `"messages"`.
	            - Prompt-only dataset: `"prompt"`.
	            - Prompt-completion dataset: `"prompt"` and `"completion"`.
	            - Preference dataset: `"prompt"`, `"chosen"`, and `"rejected"`.
	            - Preference dataset with implicit prompt: `"chosen"` and `"rejected"`.
	            - Unpaired preference dataset: `"prompt"`, `"completion"`, and `"label"`.

	        For keys `"messages"`, `"prompt"`, `"chosen"`, `"rejected"`, and `"completion"`, the values are lists of
	        messages, where each message is a dictionary with keys `"role"` and `"content"`.

	    processing_class (`ProcessingClassType`):
	        The processing_class to apply the chat template with.

	Returns:
	    `tp.Dict[str, str]`: The formatted example with the chat template applied.

	Note:
	    This function does not alter the keys, except for Language modeling dataset, where `"messages"` is replaced by
	    `"text"`.

	Example:

	>>> from transformers import AutoTokenizer
	>>> processing_class = AutoTokenizer.from_pretrained(
	...   "microsoft/Phi-3-mini-128k-instruct"
	... )
	>>> example = {
	...   "prompt": [{"role": "user", "content": "What color is the sky?"}],
	...   "completion": [{"role": "assistant", "content": "It is blue."}],
	... }
	>>> apply_chat_template(example, processing_class)
	{'prompt': '<|user|>\nWhat color is the sky?<|end|>\n<|assistant|>\n', 'completion': 'It is blue.<|end|>\n<|endoftext|>'}

	"""
	if is_conversational(example):
		return apply_chat_template(example, processing_class)
	else:
		return example


def _unpair_row(
	examples: tp.List[tp.Dict[str, tp.List[tp.Dict[str, str]]]],
) -> tp.List[tp.Dict[str, tp.List[tp.Dict[str, str]]]]:
	batch_size = len(examples["chosen"])
	new_rows = {
		"completion": examples["chosen"] + examples["rejected"],
		"label": [True] * batch_size + [False] * batch_size,
	}
	if "prompt" in examples:
		new_rows["prompt"] = examples["prompt"] + examples["prompt"]
	return new_rows


def unpair_preference_dataset(
	dataset: DatasetType, num_proc: tp.Optional[int] = None
) -> DatasetType:
	r"""
	Unpair a preference dataset.

	Args:
	    dataset (`Dataset` or `DatasetDict`):
	        Preference dataset to unpair. The dataset must have columns `"chosen"`, `"rejected"` and optionally
	        `"prompt"`.
	    num_proc (`tp.Optional[int]`, *optional*, defaults to `None`):
	        Number of processes to use for processing the dataset.

	Returns:
	    `Dataset`: The unpaired preference dataset.

	Example:

	>>> from datasets import Dataset
	>>> dataset_dict = {
	...     "prompt": ["The sky is", "The sun is"]
	...     "chosen": [" blue.", "in the sky."],
	...     "rejected": [" green.", " in the sea."]
	... }
	>>> dataset = Dataset.from_dict(dataset_dict)
	>>> dataset = unpair_preference_dataset(dataset)
	>>> dataset
	Dataset({
	    features: ['prompt', 'completion', 'label'],
	    num_rows: 4
	})
	>>> dataset[0]
	{'prompt': 'The sky is', 'completion': ' blue.', 'label': True}

	"""
	return dataset.map(
		_unpair_row,
		batched=True,
		remove_columns=["chosen", "rejected"],
		num_proc=num_proc,
	)


def maybe_unpair_preference_dataset(
	dataset: DatasetType, num_proc: tp.Optional[int] = None
) -> DatasetType:
	r"""
	Unpair a preference dataset if it is paired.

	Args:
	    dataset (`Dataset` or `DatasetDict`):
	        Preference dataset to unpair. The dataset must have columns `"chosen"`, `"rejected"` and optionally
	        `"prompt"`.
	    num_proc (`tp.Optional[int]`, *optional*, defaults to `None`):
	        Number of processes to use for processing the dataset.

	Returns:
	    `Dataset` or `DatasetDict`: The unpaired preference dataset if it was paired, otherwise the original dataset.

	Example:

	>>> from datasets import Dataset
	>>> dataset_dict = {
	...     "prompt": ["The sky is", "The sun is"]
	...     "chosen": [" blue.", "in the sky."],
	...     "rejected": [" green.", " in the sea."]
	... }
	>>> dataset = Dataset.from_dict(dataset_dict)
	>>> dataset = unpair_preference_dataset(dataset)
	>>> dataset
	Dataset({
	    features: ['prompt', 'completion', 'label'],
	    num_rows: 4
	})
	>>> dataset[0]
	{'prompt': 'The sky is', 'completion': ' blue.', 'label': True}

	"""
	from datasets import DatasetDict

	if isinstance(dataset, DatasetDict):
		column_names = dataset[list(dataset.keys())[0]].column_names
	else:
		column_names = dataset.column_names
	if "chosen" in column_names and "rejected" in column_names:
		return unpair_preference_dataset(dataset, num_proc=num_proc)
	else:
		return dataset


def extract_prompt(example: tp.Dict[str, tp.Sequence]) -> tp.Dict[str, tp.Sequence]:
	r"""
	Extracts the shared prompt from a preference data example, where the prompt is implicit within both
	the chosen and rejected completions.

	For more details, see [`maybe_extract_prompt`].
	"""
	idx = 0
	for idx in range(min(len(example["chosen"]), len(example["rejected"]))):
		if example["chosen"][idx] != example["rejected"][idx]:
			if example["chosen"][idx - 1] == " ":  # remove space before the prompt
				idx -= 1
			break
	return {
		"prompt": example["chosen"][:idx],
		"chosen": example["chosen"][idx:],
		"rejected": example["rejected"][idx:],
	}


def maybe_extract_prompt(example: tp.Dict[str, tp.List]) -> tp.Dict[str, tp.List]:
	r"""
	Extracts the shared prompt from a preference data example, where the prompt is implicit within both
	the chosen and rejected completions.

	If the example already contains a `"prompt"` key, the function returns the example as is. Else, the function
	identifies the longest common sequence (prefix) of conversation turns between the "chosen" and "rejected"
	completions and extracts this as the prompt. It then removes this prompt from the respective "chosen" and
	"rejected" completions.

	Args:
	    example (`tp.Dict[str, tp.List]`):
	        A dictionary representing a single data entry in the preference dataset. It must contain the keys
	        `"chosen"` and `"rejected"`, where each value is either conversational or standard (`str`).

	Returns:
	    `tp.Dict[str, tp.List]`: A dictionary containing:
	        - `"prompt"`: The longest common prefix between the "chosen" and "rejected" completions.
	        - `"chosen"`: The remainder of the "chosen" completion, with the prompt removed.
	        - `"rejected"`: The remainder of the "rejected" completion, with the prompt removed.

	Examples:

	>>> example = {
	...   "chosen": [
	...     {"role": "user", "content": "What color is the sky?"},
	...     {"role": "assistant", "content": "It is blue."},
	...   ],
	...   "rejected": [
	...     {"role": "user", "content": "What color is the sky?"},
	...     {"role": "assistant", "content": "It is green."},
	...   ],
	... }
	>>> extract_prompt(example)
	{'prompt': [{'role': 'user', 'content': 'What color is the sky?'}],
	 'chosen': [{'role': 'assistant', 'content': 'It is blue.'}],
	 'rejected': [{'role': 'assistant', 'content': 'It is green.'}]}

	Or, with the `map` method of `datasets.Dataset`:

	>>> from trl import extract_prompt
	>>> from datasets import Dataset
	>>> dataset_dict = {
	...   "chosen": [
	...     [
	...       {"role": "user", "content": "What color is the sky?"},
	...       {"role": "assistant", "content": "It is blue."},
	...     ],
	...     [
	...       {"role": "user", "content": "Where is the sun?"},
	...       {"role": "assistant", "content": "In the sky."},
	...     ],
	...   ],
	...   "rejected": [
	...     [
	...       {"role": "user", "content": "What color is the sky?"},
	...       {"role": "assistant", "content": "It is green."},
	...     ],
	...     [
	...       {"role": "user", "content": "Where is the sun?"},
	...       {"role": "assistant", "content": "In the sea."},
	...     ],
	...   ],
	... }
	>>> dataset = Dataset.from_dict(dataset_dict)
	>>> dataset = dataset.map(extract_prompt)
	>>> dataset[0]
	{'prompt': [{'role': 'user', 'content': 'What color is the sky?'}],
	 'chosen': [{'role': 'assistant', 'content': 'It is blue.'}],
	 'rejected': [{'role': 'assistant', 'content': 'It is green.'}]}

	"""
	# Some dataset add a `"prompt"` column, even though the prompt is implicit and included in the "chosen" and
	# "rejected" completions. E.g.:
	# {"prompt": "What color is the sky?",
	#  "chosen": [{"role": "user", "content": "What color is the sky?"}, {"role": "assistant", "content": "It is blue."}],
	#  "rejected": [{"role": "user", "content": "What color is the sky?"}, {"role": "assistant", "content": "It is green."}]}
	# That's why we check if the prompt is also conversational before deciding not to extract it.
	if "prompt" in example:
		# Both conversational or both non-conversational
		chosen_conv = is_conversational({"chosen": example["chosen"]})
		prompt_conv = is_conversational({"prompt": example["prompt"]})
		if (chosen_conv and prompt_conv) or (not chosen_conv and not prompt_conv):
			return example
	return extract_prompt({"chosen": example["chosen"], "rejected": example["rejected"]})
